/*
  dnsa -- an example of making DNS A queries using libc
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "config.h"

#ifdef HAVE_NAMESER_COMPAT
#include <arpa/nameser_compat.h>
#endif /* HAVE_NAMESER_COMPAT */

#define MAX_BUF                  4095

#define ERROR_RES_INIT           "ERROR: unable to initialize resolver library."
#define ERROR_DN_EXPAND          "ERROR: dn_expand failed for %s; this could indicate a problem with the nameserver."
#define ERROR_DN_SKIPNAME        "ERROR: dn_skipname failed for %s; this could indicate a problem with the nameserver."
#define ERROR_TYPE_A             "ERROR: dns query of %s did not include an A record as requested; this could indicate a problem with the name server."
#define ERROR_DNS_RESPONSE       "ERROR: bad or invalid dns response to %s; this could indicate a problem with the name server."

/*
 * The DNS packet format is not well documented outside of the RFCs.  There is
 * almost no sample or tutorial code to be found on the internet outside of the
 * sendmail source code, which is pretty hard to read.
 *
 * Basically, each DNS packet starts with a HEADER structure, defined in
 * arpa/nameser.h (or arpa/nameser_compat.h on Linux).  Most of the time, the
 * header can be skipped.
 *
 * After the header, the nameserver returns all of the "questions" it was asked,
 * so the answers will make sense.  If you're asking more than one "question"
 * per query, this is important.  Otherwise, skip them by finding the size of
 * each question with dn_skipname() and advancing past them.  The number of
 * questions is found in the qdcount field of the header.
 *
 * Next is the answer section, which can contain many answers, though multiple
 * answers may not make much sense for all query types.  The number of answers
 * is found in the ancount field of the header.  Within each answer, the first
 * field is the name that was queried, for reference.  It can be skipped with
 * dn_skipname().
 *
 * After that comes the type in a 16 bit field, then the class in a 16 bit
 * field, then the time-to-live (TTL) in a 32 bit field, then the answer size
 * in a 16 bit field.  The type and size are important; the class and the ttl
 * can usually be ignored.  The format of the rest of the answer field is
 * different depending on the type.
 *
 * IF THE TYPE IS A:
 * The first 4 bytes are the four octets of the IP address.
 *
 * IF THE TYPE IS T_TXT:
 * The first 8 bits are an unsigned integer indicating the total length of
 * the text response.  The following bytes are the ASCII text of the response.
 *
 * IF THE TYPE IS T_PTR OR T_NS:
 * All of the bytes are the compressed name of the result.  They can be
 * decoded with dn_expand().
 *
 * IF THE TYPE IS T_CNAME:
 * All of the bytes are the compressed name of the CNAME entry.  They can be
 * decoded with dn_expand().
 *
 * IF THE TYPE IS T_MX:
 * Each answer begins with an unsigned 16 bit integer indicating the preference
 * of the mail server (lower preferences should be contacted first).  The
 * remainder of the answer is the mail server name.  It can be decoded with
 * dn_expand().
 *
 * IF THE TYPE IS T_SOA:
 * The first section of bytes are the compressed name of the primary NS server.
 * They can be decoded with dn_expand().  The second section of bytes are the
 * compressed name of the administrator's mailbox.  They can be decoded with
 * dn_expand().  After the end of the mailbox data, five 32 bit integers give
 * the serial number, the refresh interval, the retry interval, the expiration
 * limit and the minimum time to live, in that order.
 *
 * SEE ALSO:
 *   RFC 1035
 *   http://www.zytrax.com/books/dns/ch15/
 */
int dns_a(char *target_name, int strlen_target_name)
  {
  static int initialized = 0;
  int return_value;
  int i;
  unsigned char answer[PACKETSZ];
  unsigned char host[128];
  char new_target[MAX_BUF + 1];
  unsigned char *answer_ptr;
  int strlen_new_target;
  int answer_length;
  int size;
  int type;
  int num_questions;
  int num_answers;

  return_value = 0;

  if (!initialized)
    {
    if (res_init() >= 0)
      initialized = 1;
    else
      printf(ERROR_RES_INIT);
    }

  if (initialized)
    {
    memset(answer, 0, PACKETSZ);

    if ((target_name != NULL) &&
        (strlen_target_name > 0) &&
        ((strlen_new_target = snprintf(new_target, MAX_BUF, "%.*s", strlen_target_name, target_name)) > 0) &&
        (strlen_new_target < MAX_BUF) &&
        ((answer_length = res_search(new_target, C_IN, T_A, answer, PACKETSZ)) >= 0))
      {
      answer_ptr = answer + sizeof(HEADER);

      // Skip the questions
      num_questions = ntohs((unsigned short)((HEADER *)&answer)->qdcount);
      for (i = 0; i < num_questions; i++)
        if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
          answer_ptr += size + QFIXEDSZ;
        else
          break;

      if (i == num_questions)
        {
        num_answers = ntohs((unsigned short)((HEADER *)&answer)->ancount);
        for (i = 0; (i < num_answers) && !return_value; i++)
          if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
            {
            answer_ptr += size;
            GETSHORT(type, answer_ptr);
            answer_ptr += INT16SZ; // class
            answer_ptr += INT32SZ; // ttl
            answer_ptr += INT16SZ; // size

            if (type == T_A)
              {
              printf("%d.%d.%d.%d\n", answer_ptr[0], answer_ptr[1], answer_ptr[2], answer_ptr[3]);
              answer_ptr += 4;
              }
            else if (type == T_CNAME)
              {
              if ((size = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
                {
                return_value = dns_a((char *)host, strlen((char *)host));
                answer_ptr += size;
                }
              else
                {
                printf(ERROR_DN_EXPAND, new_target);
                break;
                }
              }
            else
              {
              printf(ERROR_TYPE_A, new_target);
              break;
              }
            }
          else
            {
            printf(ERROR_DN_SKIPNAME, new_target);
            break;
            }
        }
      else
        printf(ERROR_DN_SKIPNAME, new_target);
      }
    }

  return(return_value);
  }

int main(int argc, char *argv[])
  {
  int i;

  if (argc > 1)
    for (i = 1; i < argc; i++)
      dns_a(argv[i], strlen(argv[i]));
  else
    printf(
      PACKAGE_NAME " " PACKAGE_VERSION " (C)2007 Sam Clippinger, " PACKAGE_BUGREPORT "\n"
      "http://www.spamdyke.org/\n"
      "\n"
      "USAGE: dnsa FQDN\n"
      "\n"
      "Performs a DNS lookup for the A records associated with FQDN and prints the results.\n"
      );

  return(0);
  }
